# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from six import add_metaclass
from abc import (ABCMeta, abstractmethod)
from tabulate import tabulate
from prompt_toolkit import print_formatted_text, ANSI
from collections import OrderedDict
import string
import os


__all__ = ['BoxLayout']


@add_metaclass(ABCMeta)
class BaseLayout(object):
    def __init__(self):
        pass

    @abstractmethod
    def render(self):
        pass

# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class BoxLayout(BaseLayout):
    def __init__(self, data, identifiers=None):
        super(BoxLayout, self).__init__()
        self._data = data
        self._identifiers = identifiers if identifiers else []

    def __repr__(self):
        return self.render()

    def has_children(self):
        for key, value in self._data.items():
            if isinstance(value, list) and len(value) > 0:
                return True
        return False

    def render(self):
        if not self.has_children():
            import json
            return '{}\n'.format(json.dumps(self._data, indent=4))

        identifiers = [
            {'id': 'healthStatus', 'name': 'healthy', 'style': ANSI_GREEN},
            {'id': 'healthStatus', 'name': 'unhealthy', 'style': ANSI_RED}
        ]

        for key, value in self._stack_boxes().items():
            depth = value['depth']
            if not value['parent']:
                br()
                header(
                    title={
                        'left': value['name'].capitalize() + ':' +
                        value['identifiers']['state'],
                        'right': 'Health Status: ' +
                                 value['identifiers']['healthStatus']
                       }, identifiers=identifiers
                )
            else:
                left = value['name'].capitalize()
                right = ''
                if 'identifiers' in value:
                    if left != value['parent'].capitalize():
                        left = left + ' ' + value['parent'].capitalize()
                    left = left + ':' + value['identifiers']['state']
                    right = 'Health Status: ' + \
                            value['identifiers']['healthStatus']

                Box({
                    'depth': depth,
                    'header': {
                        'left': left,
                        'right': right
                    },
                    'content': {
                        'subject': value['columns'],
                        'body': value['rows']
                    },
                    'identifiers': identifiers
                }).render()
                br()
        return ''

    def _stack_boxes(self):
        def stack(data, boxes, parent='', name='', trail='', identifiers=None):
            empty_slot = '-'
            for key, value in data.items():
                # assume name is first XYXName
                if isinstance(key, str) and 'Name' in key:
                    name = value

                if not parent and not trail:
                    trail = name
                    boxes[trail] = OrderedDict([
                        ('name', name),
                        ('parent', parent),
                        ('identifiers', OrderedDict()),
                        ('depth', len(trail.split('.')) - 1),
                        ('columns', list(data.keys())[:-1]),
                        ('rows', [[]])
                    ])
                if isinstance(value, dict):
                    # lookahead for list if no list, stop
                    contains_list = False
                    rows = []
                    for d_k, d_v in value.items():
                        rows.append([d_k, d_v if d_v else empty_slot])
                        if isinstance(d_v, list):
                            contains_list = True
                    if contains_list:
                        stack(value, boxes, parent, name, trail, identifiers)
                    else:
                        # -- generic box structure --
                        t = trail + '.' + key
                        boxes[t] = OrderedDict([
                            ('name', key),
                            ('parent', parent),
                            ('depth', len(t.split('.')) - 1),
                            ('columns', ['name', 'url']),
                            ('rows', rows)
                        ])

                elif isinstance(value, list):
                    new_empty_box = False
                    for box_item in value:
                        if not new_empty_box:
                            new_empty_box = True

                            trail = name + ':' + trail + '.' + key if parent \
                                else name + ':' + key

                            # -- generic box structure --
                            boxes[trail] = OrderedDict([
                                ('name', name if parent else key),
                                ('parent', parent if parent else key),
                                ('identifiers', OrderedDict()),
                                ('depth', len(trail.split('.')) - 1),
                                ('columns', list(box_item.keys())[:-1]),
                                ('rows', [])
                            ])
                            parent = key

                        # -- seed new empty row slot to row context of box --
                        box = boxes[trail]
                        rows = box['rows']
                        rows.append([])
                        box['rows'] = rows
                        boxes[trail] = box

                        stack(box_item, boxes, parent, name, trail, identifiers)
                else:
                    # -- add column value to current row context of box --
                    if trail in boxes:
                        box = boxes[trail]

                        # To remain generic use use identifier keys to stash any
                        # context values by the identifying key(s)
                        if identifiers:
                            for identify in identifiers:
                                if identify == key:
                                    content = box['identifiers']
                                    content[key] = ' ' + value + ' '
                                    box['identifiers'] = content
                                    break

                        if key in box['columns']:
                            rows = box['rows']
                            # current row context is the last row
                            last_row = rows[-1]
                            # add column value to current row context
                            last_row.append(value if value else empty_slot)
                            # persist row changes back to the current context
                            rows[-1] = last_row
                            box['rows'] = rows
                            boxes[trail] = box

        stacked_boxes = OrderedDict()
        stack(self._data, stacked_boxes, identifiers=self._identifiers)

        return stacked_boxes


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class Box(object):
    def __init__(self, data):
        self._data = data
        self._identifiers = data.get('identifiers', []) if \
            'identifiers' in data else []

    def render(self):
        data = self._data
        identifiers = self._identifiers

        # -- Render header --
        header(
            title=data['header'],
            char='-',
            depth=data['depth'],
            identifiers=identifiers,
        )

        # -- Render table content --
        content = data['content']
        columns = [string.capwords(s) + '\n' for s in content['subject']]

        rows = content['body']
        for row in tabulate(rows, columns, 'plain').split('\n'):
            badge(row, depth=data['depth'], end='\n', identifiers=identifiers)


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #

ANSI_GREEN = 102
ANSI_RED = 101


def br():
    print('\n')


def indention(depth):
    padding = []
    margin = 3
    offset = margin * depth
    offset = 1 if offset <= margin else offset

    for i in range(0, offset):
        padding.append(' ')

    return ''.join(padding)


def badge(text, depth=0, end='', identifiers=None):
    def background(label, color):
        print_formatted_text(
            ANSI('\x1b[97;{color}m{label}'.format(
                color=color, label=label)), end='')

    padding = indention(depth)

    if identifiers:
        for identifier in identifiers:
            tokens = text.split(' ' + identifier['name'] + ' ')
            if len(tokens) == 2:
                print_formatted_text(padding + tokens[0] + ' ', end='')
                background(identifier['name'], identifier['style'])
                print_formatted_text(' ' + tokens[1], end=end)
                return

    print_formatted_text(padding + text, end=end)


def div(char='-', depth=0):
    w = terminal_width()
    margin_left = indention(depth)
    divider = [margin_left]
    index = margin_left.count('')
    while index < w:
        divider.append(char)
        index = index + 1
    print_formatted_text('')
    print_formatted_text(''.join(divider))


def header(title, char='=', depth=0, identifiers=None):
    title_left = title.get('left', '')
    title_right = title.get('right', '')

    w = terminal_width()
    margin_left = indention(depth)
    index = len(margin_left) + 1
    offset = w - len(margin_left) - len(title_left) - len(title_right)
    padding = []
    while index <= offset:
        padding.append(' ')
        index = index + 1

    label = '{}{}'.format(margin_left, title_left)
    print_formatted_text(label, end='')
    if title_right:
        badge(''.join(padding) + title_right, depth=depth, end='',
              identifiers=identifiers)
    div(char=char, depth=depth)


def terminal_width():
    try:
        width = os.get_terminal_size().columns
    except OSError:
        import sys
        width = sys.maxsize

    return width
